;; Generated package description from dirvish.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "dirvish" "2.3.0.0.20250504.80741" "A modern file manager based on dired mode" '((emacs "28.1") (compat "30")) :commit "d877433f957a363ad78b228e13a8e5215f2d6593" :authors '(("Alex Lu" . "https://github.com/alexluigit")) :maintainer '("Alex Lu" . "https://github.com/alexluigit") :keywords '("files" "convenience") :url "https://github.com/alexluigit/dirvish")
